(****************************************************************************)
(* An implementation of Higher-Order Pattern Unification in SML             *)
(* Copyright (C) 2004 Gopalan Nadathur and Natalie Linnell                  *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(*
   This file contains code for realizing a representation of lambda terms
   based on the suspension notation of Nadathur and Wilson, adapted to a 
   situation where terms appear under a mixed prefix of quantifiers. 
   The components defined are

         TERM   a signature for terms
         Term   a structure that concretely realizes TERM

   Some specific comments concerning the representation of terms:
   Constants (const) correspond to universally quantified variables and 
   they carry with them a name and the number of changes from existential 
   to universal quantifiers prior to their quantifier. Existential variables 
   (var) carry similar information. A series of abstractions are bunched 
   together, e.g. lam lam t is represented as lam(2,t). For applications,
   a first-order like representation is used; e.g. (t t_1 ... t_n) is 
   represented as (t [t_1,...,t_n]). Finally terms are pointers to 
   appropriate structures to support destructive updates both in unification
   and in normalization. 
*)

signature TERM =
sig

datatype rawterm =  const of string * int
                 |  var of string * int 
                 |  dB of int
                 |  lam of int * (rawterm ref)
                 |  app of (rawterm ref) * (rawterm ref) list
                 |  susp of (rawterm ref) * int * int * (envitem list)
                 |  ptr of (rawterm ref)
and envitem = dum of int
            | bndg of (rawterm ref) * int

type term = (rawterm ref)
type env = (envitem list)

val deref : term -> term
val nth : env*int -> envitem
val add_dummies : env -> int -> int -> env
val makelambda : int -> term -> term
val bind : term -> term -> unit
val freshvar : int -> term

end




structure Term : TERM =
struct

val varcount = (ref 0)

datatype rawterm =  const of string * int
                 |  var of string * int 
                 |  dB of int
                 |  lam of int * (rawterm ref)
                 |  app of (rawterm ref) * (rawterm ref) list
                 |  susp of (rawterm ref) * int * int * (envitem list)
                 |  ptr of (rawterm ref)
and envitem = dum of int
            | bndg of (rawterm ref) * int

type term = (rawterm ref)

type env = (envitem list)

fun deref (ref(ptr(t))) = deref(t)
  | deref term = term

fun nth(x::l,1) = x
  | nth(x::l,n) = nth(l,n-1)

fun add_dummies e 0 _ = e
  | add_dummies e n nl = 
       let val n' = (n-1) 
       in ((dum (nl+n')) :: (add_dummies e n' nl))
       end

(* Construct an abstraction from the no of abstractions and the body *)
fun makelambda 0 t = t
  | makelambda n (ref (lam(n',t))) = (ref (lam (n+n',t)))
  | makelambda n t = (ref (lam(n,t)))


(* Binding a variable to a term. The *contents* of the cell the
   variable is a reference to must be updated. Also the variable must
   not be made a reference to itself. This can be changed to mimic the
   Prolog representation of bound variables but then deref will have to
   work differently. This is the place to introduce trailing.
*)
fun bind v t = if (v = t) then () else (v := ptr(t))


(* Generating a fresh variable with a given time stamp; the use of ref
   ensures uniqueness. We should attach useful names as well, but this 
   will do for the moment.
*)
fun freshvar ts2 = 
      let val name = concat(["h", Int.toString(!varcount)])
      in
	varcount := !varcount + 1 ;
        ref (var (name,ts2))
      end

end

